C     Collection of routines for applying 2D Helmholtz integral
C     operators S, D, D^T, and T, to densities on a smooth closed curve segment
C     defined by a set of quadrature points (periodic trapezoid rule with
C     respect to the parametrization of the curve on [0 2pi) ), with FMM
C     acceleration and Alpert local corrections.
C
C     First part of a supposed 2D layer-potential toolbox, lp2d, containing
C     self-interactions of open and close segments, simple far-evaluation,
C     close-evaluation, corners...
C
C     Uses Fortran 90 for dynamic array allocation.
C     Use makefile to compile, including driver lpevalselfcc_dr into lp2d.
C     Use makefile.mwrap for creating MEX interfaces for MATLAB.
C
C    Alex Barnett 3/19/11.  Some code from Mike O'Neil & Zydrunas Gimbutas.
C
C     Remarks:
C     1) Some effort has been spent making the Alpert correction fast.
C        At 16th order they take around 40% of the time to apply the FMM.
C     2) Alpert corrections do not work for the hypersingular operator T,
C        but do work for the difference of two such operators at different
C        wavenumbers on the same curve. This introduces round-off error
C        that we will live with for now.
C     3) There are further corrections for quasi-periodic open segments. We
C        don't include these yet.


C============================== main interface ==============================
C
      subroutine lpevalselfcc(ier, iprec, n, node,nvec,speed, ifslp,
     $     slpdens,
     $     ifdlp,dlpdens, zk, zk2, nord, qpwinding,qpdvec,qpalpha,qpnei,
     $     ifpot,pot, ifnd,nd)
C     Self-evaluation of (Helmholtz) layer-potentials on a 2D closed curve.
C     Jump relations are not included, just applying the integral operators.
C
C     INPUTS:
C     iprec     : precision for FMM (integer) - see HFMM2D documentation
C     n         : number of quadrature nodes on curve
C     node(2,n) : quadrature node x,y locations defining closed curve
C     nvec(2,n) : unit normal n_x,n_y components at each node
C     speed(n)  : size of derivative of parametrization, sqrt(x'(t)^2+y'(t)^2)
C                 at each node
C     ifslp     : flag for presence of single-layer potential.
C                   0 no SLP, 1 is SLP
C     slpdens(n): list of SLP density function value at the nodes (complex *16)
C     ifdlp     : flag for presence of double-layer potential.
C                   0 no DLP, 1 is DLP
C     dlpdens(n): list of DLP density function value at the nodes (complex *16)
C     zk        : wavenumber (complex *16)
C     zk2       : difference wavenumber (complex *16), interpreted as follows:
C                 if zk2 on negative real axis, apply operator Op_zk only
C                   at wavenumber zk,
C                 otherwise, apply difference operator Op_zk - Op_zk2
C     nord      : order for Alpert quadrature (integer),
C                 0 gives no correction; otherwise must be 4, 8, or 16.
C     qpwinding : if 0, segment treated as closed in the plane
C                 if 1, segment is treated as open but closed in the periodic
C                 cylinder, going left-to-right.
C                 if -1, ditto, going right-to-left.
C     qpdvec(2) : displacement vector for quasi-periodicity (rightwards-going)
C     qpalpha   : Bloch phase for quasi-periodicity (complex *16)
C     qpnei     : 0,1,2... controls number of neighbor images summed (integer)
C                 qpnei=0 switches off QP features; qpnei>0 not implemented yet.
C     ifpot     : flag (0 or 1) controlling if potential output is computed
C     ifnd      : flag (0 or 1) controlling if normal deriv output is computed
C
C     OUTPUTS:
C     pot(n)    : potential values at the nodes (complex *16)
C     nd(n)     : normal-derivative values at the nodes (complex *16)
C     ier       : error return code. 0 is ok.
C                 ier=8 - failure to dynamically allocate workspace
C
C     Comments:
C     1) jump relations not included, since depends on limiting side to curve.
C     2) x,y,n_x,n_y,speed are interpolated when needed in between nodes.
C     3) Parameterization is t in [0,2pi). Nodes are uniform with respect to t.
C     4) zk2 (not yet implemented) is so that T_1-T_2 can be computed without
C        round-off error problems.

c     I/O vars:
      implicit none
      integer ier,iprec,n, ifslp, ifdlp, nord, ifpot, ifnd
      integer qpnei, qpwinding
      real *8 node(2,n), nvec(2,n), speed(n), qpdvec(2)
      complex *16 slpdens(n), dlpdens(n), zk, zk2, qpalpha, pot(n)
      complex *16 nd(n)
c     internal working vars:
      integer i, j, optype, ifhess, neiind, nein, ifpotself, ifndself
      real *8 pi, w
      complex *16 ima, par, hess(3), phase
      parameter (pi = 3.141592653589793d0, ima = (0, 1))
c     workspace (f90 dynamic allocation)...
      complex *16, allocatable :: a(:), b(:), grad(:,:)
      real *8, allocatable :: neinode(:,:), neinvec(:,:)
      complex *16, allocatable :: neia(:), neib(:)
      external fourkernels

c     verb: first arg = 6 writes stderr, 0 not; 2nd arg = 13 writes to fort.13
      call prini(6,13)

      allocate(a(n), b(n), grad(2,n), stat=ier)
      if (ier.ne.0) then
         call prinf('failure to allocate, with n=*',n,1)
         ier = 8
         return
      endif

c     dummy kernel param for now...
      par = (0,0)
c     regular node quadr weights in the t parametrization variable
      w = 2d0*pi/n
c     fill a=charge and b=dipole strengths...
      if (ifslp.eq.1) then
         do i=1,n
            a(i) = w * speed(i) * slpdens(i)
         enddo
      endif
      if (ifdlp.eq.1) then
         do i=1,n
            b(i) = w * speed(i) * dlpdens(i)
         enddo
      endif

c     Do FMM self-eval, then normal-derivatives = dot(grad,curvenormals)...
      ifhess = 0
c      call prin2('a=*',a,10)
c      call prin2('b=*',b,10)
      call prini(0,13)
      call hfmm2dpart(ier, iprec, zk, n, node, ifslp, a,
     $     ifdlp, b, nvec, ifpot, pot, ifnd, grad, ifhess, hess)
      call prini(6,13)
      if (ier.ne.0) then
         call prinf('problem in hfmm2d, with n=*',n,1)
         return
      endif
c     remark: maybe could avoid allocating some of grad,a,b if FMM can use
c     same output space as input space for charge, dipole. Who cares?
      if (ifnd.eq.1) then
         do i=1,n
            nd(i) = nvec(1,i)*grad(1,i) + nvec(2,i)*grad(2,i)
         enddo
      endif

c     Sum the nearest neighbors if needed... (using RAM-intensive way)
      if (qpnei.gt.0) then
c         deallocate(a)
         nein = n*2*qpnei
         allocate(neia(nein), neib(nein), neinode(2,nein),
     $        neinvec(2,nein), stat=ier)
         if (ier.ne.0) then
            call prinf('failure to allocate, with nein=*',nein,1)
            ier = 9
            return
         endif         
c     fill neia=charge, neib=dipole strengths (j loops over copies)
         neiind = 0
         do j=-qpnei,qpnei
            if (j.ne.0) then
               phase = qpalpha ** j
               do i=1,n
                  neinode(1,i+neiind) = node(1,i) + qpdvec(1)*j
                  neinode(2,i+neiind) = node(2,i) + qpdvec(2)*j
               enddo
               if (ifslp.eq.1) then
                  do i=1,n
                     neia(i+neiind) = phase * w * speed(i) * slpdens(i)
                  enddo
               endif
               if (ifdlp.eq.1) then
                  do i=1,n
                     neinvec(1,i+neiind) = nvec(1,i)
                     neinvec(2,i+neiind) = nvec(2,i)
                     neib(i+neiind) = phase * w * speed(i) * dlpdens(i)
                  enddo
               endif
               neiind = neiind + n
            endif
         enddo         
c     no self-eval of these sources (hess is dummy); a has targ pot
         ifpotself = 0
         ifndself = 0
         call prini(0,13)
         call hfmm2dparttarg(ier, iprec, zk, nein,neinode, ifslp,
     $        neia, ifdlp, neib, neinvec, ifpotself, hess, ifndself,
     $        hess, ifhess, hess, n, node, ifpot,a,ifnd,grad,ifhess,
     $        hess)
         call prini(6,13)
         if (ier.ne.0) then
            call prinf('problem in hfmm2d, with n=*',n,1)
            return
         endif
c     add these to the pot and nd output arrays if necessary
         if (ifpot.eq.1) then
            do i=1,n
               pot(i) = pot(i) + a(i)
            enddo
         endif
         if (ifnd.eq.1) then
            do i=1,n
               nd(i) = nd(i) + nvec(1,i)*grad(1,i) + nvec(2,i)*grad(2,i)
            enddo
         endif
      endif

c     Add Alpert corrections for 2x2 block of operators at once...
c     (For case of >1 operators needed, this is faster than doing one by one)
      call applyalpertcc(ier, n, node, nvec, speed, zk, zk2,
     $     par, fourkernels, nord, ifslp, slpdens, ifdlp, dlpdens,
     $     ifpot, pot, ifnd, nd)

      return
      end


C ================ IDEA FOR MULTIPLE CC CASE ================================
      subroutine lpevalselfmanycc(ier, iprec, n, node,nvec,speed,
     $     nccs, 
     $     ifslp,slpdens,
     $     ifdlp,dlpdens, zk, zk2, nord, qpwinding,qpdvec,qpalpha,qpnei,
     $     ifpot,pot, ifnd,nd)
C     Self-evaluation of (Helmholtz) layer-potentials between many 2D closed
C     curves at the same wavenumber.
C     Jump relations are not included, just applying the integral operators.
C
C     nccs     : number of isolated closed curves
C     cclens   : numbers of nodes in each of the 1,...,nccs closed curves

C   Type of each closed curve? closed, open Gaussian, etc. ?

C    qpwinding : list of winding numbers for 1,..., nccs (integer)

C     zk2      : list of complex * 16 wavenumbers for backsides of each
C                closed curve. If a zk2 is on negative Re axis, don't subtract
C                operators.

C Alg: do one big FMM. Then break Alpert down into chunks for each cc, using
C    applyalpertcc on each.
C    For open QP ones use twaeked QP Alpert...

C 3/21/11

c     I/O vars:
      implicit none
      integer ier,iprec,n, ifslp, ifdlp, nord, ifpot, ifnd
      integer qpnei, qpwinding, nccs, cclens(nccs)
      real *8 node(2,n), nvec(2,n), speed(n), qpdvec(2)
      complex *16 slpdens(n), dlpdens(n), zk, zk2(nccs), qpalpha, pot(n)
      complex *16 nd(n)

      return
      end



c****************************************************************************

      subroutine applyalpertcc(ier, n, node, nvec, speed, par1,
     $     par2, par3, kerfun, norder, ifslp, slpdens, ifdlp, dlpdens,
     $     ifpot, pot, ifnd, nd)
c     Adds Alpert correction at all nodes, given the density values at
c     these nodes on a closed curve (cc). Is agnostic as to kernel function.
c     (Thus, it can be used to apply S, D, D^T or T operators, or others...)
c
c     If you want to apply correction for a single density hit by a single
c     kernel: send in density as slpdens, have kerfun return a single value (f),
c     set ifslp=1, ifdlp=0, ifpot=1, ifnd=0; the corr. will be applied to pot.
c     If you want to correct more than one operator in the 2x2 block at once,
c     choose appropriate flags, and have kernel return a list of 4 values (f).
c     This enables S, D, D^T and T to be corrected in an efficient single call.
c
c     The near-diagonal regular node interactions are first subtracted, then
c     the effect of the Alpert extra nodes and weights is added in.
c     Typical input arrays in pot and nd would be potential and n-derivs at
c     the nodes produced by an FMM call to the (quadrature-weighted) density,
c     as in lpevalselfcc. (The latter FMM call should omit self-interaction of
c     nodes, obviously).
c
c     INPUTS:
C     n         : number of quadrature nodes on curve
C     node(2,n) : quadrature node x(t),y(t) locations defining closed curve
C     nvec(2,n) : unit normal x,y components at each node
C     speed(n)  : size of derivative of parametrization, sqrt(x'(t)^2+y'(t)^2)
C                 at each node
c     par1,par2,par3 : parameters passed to kernel function,
c                        eg wavenumber(s), etc (complex *16)
c     kerfun    : kernel function, returning value of kernel without the speed
c                 factor. Must have calling sequence as follows...
c
c          kerfun(src, snor, targ, tnor, par1, par2, par3, f)
c          kerfun inputs:
c                 src(2)    - source location 1- and 2-coordinate in R^2
c                 snor(2)   - unit normal at source, 1- and 2-coordinate
c                 targ(2)   - target location 1- and 2-coordinate
c                 tnor(2)   - unit normal at target, 1- and 2-coordinate
c                 par1,par2,par3 - parameters for kernel (eg wavenumber, etc)
C                                    (complex*16)
c          kerfun output:   
c                 f         - value(s) of kernel function(s) (complex *16).
c                             If ifdlp=0 and ifnd=0, f can be a single value;
c                             otherwise f must return the list of 4 values
c                             in the 2x2 block,
c                        [pot; nd] = [f(1) f(2); f(3) f(4)] [slpdens; dlpdens]
c
c     norder    : Alpert order (4,8, or 16). norder=0 results in u=0 output.
c     ifslp     : slpdens computation flag
c                      ifslp = 0 -> don't include slpdens; 1 -> do.
C     slpdens(n): 1st (eg SLP) density function at the nodes (complex *16)
c     ifdlp     : dlpdens computation flag
c                      ifdlp = 0 -> don't include dlpdens; 1 -> do.
C     dlpdens(n): 2nd (eg DLP) density function at the nodes (complex *16)
c     ifpot     : flag controlling potential correction
c                      ifpot = 0 -> don't add corr. to pot; 1 -> do.
c     ifnd      : flag controlling normal-derivative correction
c                      ifnd = 0 -> don't add corr. to nd; 1 -> do.
c
c     OUTPUTS:
c     ier       : 0 is ok, otherwise an error. These include:
c                 1 is unknown Alpert norder
c
c     INPUT/OUTPUTS:
c     pot(n)    : vector of values at the nodes (complex *16)
c     nd(n)     : vector of normal-derivatives at nodes (complex *16)
c
c     Notes:
c     1) pot (if ifpot=1) and nd (if ifnd=1) must already be allocated.
c     2) algorithm is O(p^2 n) where p is Alpert order. This is achieved by
c        prestoring the interpolation coeffs so they are not computed for
c        each Alpert node for each target point. Kernel called about 3pn times,
c        which can be about 50n - this starts to approach the FMM effort
c        itself, strangely. But you can't really do better than that.
c     3) The use of flags ifslp,ifdlp,ipot,ifnd are an attempt to keep this
c        code kernel-independent, but still have maximum efficiency when
c        correcting a 2x2 operator block that
c        shares similar kernels (hence should be evaluated in 1 kerfun call).
c
c     Alex Barnett 3/18/11
      implicit none
      integer n, norder, ier, nextra, nskip, i, j, jspill, nint, l, k
      integer ifslp, ifdlp, ifpot, ifnd
      real *8 node(2,n), nvec(2,n), speed(n), w, c
      complex *16 pot(n), slpdens(n), f(4), par1, par2, par3
      complex *16 nd(n), dlpdens(n), sspdensl, dspdensl
      real *8 extranodes4(6), extraweights4(6)
      real *8 extranodes8(14), extraweights8(14)
      real *8 extranodes16(30), extraweights16(30)
      real *8 extranodes(30), extraweights(30)
      real *8 src(2), targ(2), snor(2), tnor(2), bwei(30,30), coef(30)
      real *8 rboff(30)
      integer boff(30,30)
c     functions
      external kerfun
      integer nmod
c     math constants
      real *8 pi
      parameter (pi = 3.141592653589793d0)

c---------- quadratures from Mike O'Neil's code test5.f:
      data extranodes4/
     1 -1.023715124251890D+00,
     1 -2.935370741501914D-01,
     1 -2.379647284118974D-02, 
     1 2.379647284118974D-02, 
     1 2.935370741501914D-01,
     1 1.023715124251890D+00/ 
      data extraweights4/
     1 9.131388579526912D-01,
     1 4.989017152913699D-01,
     1 8.795942675593887D-02,
     1 8.795942675593887D-02,
     1 4.989017152913699D-01,
     1 9.131388579526912D-01/
c
      data extranodes8/
     1 -3.998861349951123D+00,
     1 -2.980147933889640D+00,
     1 -1.945288592909266D+00,
     1 -1.027856640525646D+00,
     1 -3.967966533375878D-01,
     1 -9.086744584657729D-02,
     1 -6.531815708567918D-03,
     1 6.531815708567918D-03,
     1 9.086744584657729D-02,
     1 3.967966533375878D-01,
     1 1.027856640525646D+00,
     1 1.945288592909266D+00,
     1 2.980147933889640D+00,
     1 3.998861349951123D+00/
      data extraweights8/
     1 1.004787656533285D+00,
     1 1.036093649726216D+00,
     1 1.008710414337933D+00,
     1 7.947291148621894D-01,
     1 4.609256358650077D-01,
     1 1.701315866854178D-01,
     1 2.462194198995203D-02,
     1 2.462194198995203D-02,
     1 1.701315866854178D-01,
     1 4.609256358650077D-01,
     1 7.947291148621894D-01,
     1 1.008710414337933D+00,
     1 1.036093649726216D+00,
     1 1.004787656533285D+00/
c
      data extranodes16/
     1 -8.999998754306120d+00,
     1 -7.999888757524622d+00,
     1 -6.997957704791519d+00,
     1 -5.986360113977494d+00,
     1 -4.957203086563112d+00,
     1 -3.928129252248612d+00,
     1 -2.947904939031494d+00,
     1 -2.073471660264395d+00,
     1 -1.348993882467059d+00,
     1 -7.964747731112430d-01,
     1 -4.142832599028031d-01,
     1 -1.805991249601928d-01,
     1 -6.009290785739468d-02,
     1 -1.239382725542637d-02,
     1 -8.371529832014113d-04,
     1 8.371529832014113d-04,
     1 1.239382725542637d-02,
     1 6.009290785739468d-02,
     1 1.805991249601928d-01,
     1 4.142832599028031d-01,
     1 7.964747731112430d-01,
     1 1.348993882467059d+00,
     1 2.073471660264395d+00,
     1 2.947904939031494d+00,
     1 3.928129252248612d+00,
     1 4.957203086563112d+00,
     1 5.986360113977494d+00,
     1 6.997957704791519d+00,
     1 7.999888757524622d+00,
     1 8.999998754306120d+00/
      data extraweights16/
     1 1.000007149422537d+00,
     1 1.000395017352309d+00,
     1 1.004798397441514d+00,
     1 1.020308624984610d+00,
     1 1.035167721053657d+00,
     1 1.014359775369075d+00,
     1 9.362411945698647d-01,
     1 8.051212946181061d-01,
     1 6.401489637096768d-01,
     1 4.652220834914617d-01,
     1 3.029123478511309d-01,
     1 1.704889420286369d-01,
     1 7.740135521653088d-02,
     1 2.423621380426338d-02,
     1 3.190919086626234d-03,
     1 3.190919086626234d-03,
     1 2.423621380426338d-02,
     1 7.740135521653088d-02,
     1 1.704889420286369d-01,
     1 3.029123478511309d-01,
     1 4.652220834914617d-01,
     1 6.401489637096768d-01,
     1 8.051212946181061d-01,
     1 9.362411945698647d-01,
     1 1.014359775369075d+00,
     1 1.035167721053657d+00,
     1 1.020308624984610d+00,
     1 1.004798397441514d+00,
     1 1.000395017352309d+00,
     1 1.000007149422537d+00/
c
c     set up the Alpert extra nodes and weights arrays (Mike O'Neil)...
      if (norder.eq. 0) then
         nskip=1
         nextra=0
      elseif (norder.eq.4) then
         nskip = 2
         nextra = 6
	 do i = 1,nextra
	    extranodes(i) = extranodes4(i)
	    extraweights(i) = extraweights4(i)
         enddo
      elseif (norder.eq.8) then 
         nskip = 5
         nextra = 14
	 do i = 1,nextra
	    extranodes(i) = extranodes8(i)
	    extraweights(i) = extraweights8(i)
         enddo
      elseif (norder.eq.16) then 
         nskip = 10
         nextra = 30
	 do i = 1,nextra
	    extranodes(i) = extranodes16(i)
	    extraweights(i) = extraweights16(i)
         enddo
      else
         ier = 1
         call prinf('wrong order for Alpert, norder=*',norder,1)
c        don't use stop since it kicks you right out of Matlab!
         return
      endif      
      ier = 0

c     regular node quadr weights in the t parametrization variable
      w = 2d0*pi/n

c     initialize f on the off-chance that your kerfun doesn't provide f(2:4)
      do i=1,4
         f(i) = 0
      enddo

c     Subtract off effect of regular nodes closer than nskip to ith target...
      do i=1,n
         targ(1) = node(1,i)
         targ(2) = node(2,i)
         tnor(1) = nvec(1,i)
         tnor(2) = nvec(2,i)
c     loop over near enough sources j (wrapping mod n, and skipping i itself)
         do jspill = i-nskip+1,i+nskip-1
            if (jspill.ne.i) then
               j = nmod(jspill,n);
               src(1) = node(1,j)
               src(2) = node(2,j)
               snor(1) = nvec(1,j)
               snor(2) = nvec(2,j)
               call kerfun(src,snor,targ,tnor,par1,par2,par3,f)
               c = -w*speed(j)
c     add to output using kernel value(s) f times densities, with factor c
               if (ifpot.eq.1) then
                  if (ifslp.eq.1) then
                     pot(i) = pot(i) + c * slpdens(j) * f(1)
                  endif
                  if (ifdlp.eq.1) then
                     pot(i) = pot(i) + c * dlpdens(j) * f(2)
                  endif
               endif
               if (ifnd.eq.1) then
                  if (ifslp.eq.1) then
                     nd(i) = nd(i) + c * slpdens(j) * f(3)
                  endif
                  if (ifdlp.eq.1) then
                     nd(i) = nd(i) + c * dlpdens(j) * f(4)
                  endif
               endif
            endif
         enddo
      enddo

c     interpolation order (found by experiment with alpertizeselfmatrix.m)
      nint = norder+3

c     Fill arrays: node offsets (boff) and interpolation coeffs (bwei)
c     for each extra node. boff(j,l) is coef of jth offset regular node to
c     value at lth extranode...
      do l=1,nextra
c     set boff col to nearest nint reg node index offsets for this extra node
        boff(1,l) = floor(extranodes(l)-nint/2+1)
        do j=1,nint
           boff(j,l) = boff(1,l) + j - 1
c          note bary needs a real*8 version (rboff) of the lth col of boff...
           rboff(j) = boff(j,l)
        enddo
c     send lth col of boff as the interp nodes, dump coefs into lth bwei col
        call bary1_coefs(nint, rboff, extranodes(l), bwei(1,l))
      enddo
c     Remark: I don't use h-scaling, since it could get small (1e-6), bad
c     for bary weights underflow!

c     Use interpolation arrays for Alpert corrections on each ith target...
      do i=1,n
         targ(1) = node(1,i)
         targ(2) = node(2,i)
         tnor(1) = nvec(1,i)
         tnor(2) = nvec(2,i)
         do l=1,nextra
c     initialize interpolant values at lth extranode, then fill them...
            sspdensl = (0,0)
            dspdensl = (0,0)
            src(1) = 0d0
            src(2) = 0d0
            snor(1) = 0d0
            snor(2) = 0d0
            do k=1,nint
c     from offset get wrapped index of kth interp node
               j = nmod(i+boff(k,l), n)
               c = bwei(k,l)
c     speed * density... (uses flops but saves another n array)
c     Lame that if statements happen O(p^2n) times, but doesn't affect speed!
c     (needed to prevent reading from invalid density array)
               if (ifslp.eq.1) then
                  sspdensl = sspdensl + c*speed(j)*slpdens(j)
               endif
               if (ifdlp.eq.1) then
                  dspdensl = dspdensl + c*speed(j)*dlpdens(j)
               endif
               src(1) = src(1) + c*node(1,j)
               src(2) = src(2) + c*node(2,j)
               snor(1) = snor(1) + c*nvec(1,j)
               snor(2) = snor(2) + c*nvec(2,j)
            enddo
c     eval exact kernel val f at interpolated source pts & normal
            call kerfun(src,snor,targ,tnor,par1,par2,par3,f)
            c = w * extraweights(l)
c     add to output using kernel value(s) f times densities, with factor c
            if (ifpot.eq.1) then
               if (ifslp.eq.1) then
                  pot(i) = pot(i) + c * sspdensl * f(1)
               endif
               if (ifdlp.eq.1) then
                  pot(i) = pot(i) + c * dspdensl * f(2)
               endif
            endif
            if (ifnd.eq.1) then
               if (ifslp.eq.1) then
                  nd(i) = nd(i) + c * sspdensl * f(3)
               endif
               if (ifdlp.eq.1) then
                  nd(i) = nd(i) + c * dspdensl * f(4)
               endif
            endif
         enddo
      enddo

      return
      end


c**********************KERNELS EVALUATION INTERFACE****************************
      subroutine fourkernels(src, snor, targ, tnor, zk, zk2, par, f)
c     Evaluate the kernels of
c     the four Helmholtz operators S, D, D^T and T simultaneously, between
c     one source (with unit normal) and one target point (with unit normal),
c     both in R^2.
C     The point of this routine is to make only *one* call to hank103 (which
c     happens in hpotgrad2d_sdp2).
c     The kernel of Helmholtz S operator from one source to one target point:
c         f(1) = (i/4) H_0^{(1)}(k|x-y|)
C                      where, y = source, n_y = unit normal at source
C                             x = target, n_x = unit normal at target.
C     Then f(2) through f(4) are D, D^T and T, with the same normalization.
c
c          Inputs:
c                 src(2)   - source location 1- and 2-coordinate in R^2
c                 snor(2)  - unit normal at source, 1- and 2-coordinate
c                 targ(2)  - target location 1- and 2-coordinate
c                 tnor(2)  - unit normal at target, 1- and 2-coordinate
c                 zk       - wavenumber (complex *16)
c                 zk2      - difference wavenumber (not yet used) (complex *16)
c                 par      - extra parameter (not yet used) (complex *16)
c          Output:   
c                 f(4)     - values of kernels (complex *16) in the order:
c                            S, D, D^T (ie S'), T (ie D').
c Remarks:
c 1) Future plan is to use zk2, when not NaN, to return *difference* of kernel
c    at wavenumbers zk and zk2. This would allow accurate Taylor expansion for
c    small arguments as needed by the close Alpert extra nodes, esp. for T op.
      implicit none
      real *8 src(2), snor(2), targ(2), tnor(2)
      complex *16 zk, zk2, f(4), grad(2), grad_dp(2), hess(3), par
c     Tweaked Gimbutas routine, writes into f(1) and f(2); hess is dummy...
      call hpotgrad2d_sdp2(zk,src,snor,targ,1,f(1),f(2),1,grad,grad_dp,
     $     0,hess,hess)
c     n-derivatives at target...
      f(3) = tnor(1)*grad(1) + tnor(2)*grad(2)
      f(4) = tnor(1)*grad_dp(1) + tnor(2)*grad_dp(2)
      return
      end



C******************************** UTILITIES *********************************

      integer function nmod(i, n)
C     returns 1-indexed modulus of i, ie wrapped into [1,n]. Handles i<0
c     correctly. Barnett 3/15/11
      integer n, i, j
      j = mod(i-1,n)
      if (j.lt.0) then
         j = j + n
      endif
      nmod = j+1
      return
      end

      subroutine bary1_coefs(n,ts,ttt,coefs)
      implicit real *8 (a-h,o-z)
      real *8 ts(1),ttt,whts(1000),coefs(1)
c     
c     Use barycentric interpolation on nodes ts to evaluate at ttt,
c     returning the coefficients with which to weight function values f(ts)
c     to get the interpolant at ttt
c
c       input:
c         n - the length of ts,xs
c         ts - nodes with which to interpolate
c         ttt - point at which to interpolate
c
c       output:
c         coefs - coefficients for the interpolation
c
c     (Code from Mike O'Neil, from test15.f, 3/15/11)

      do 1200 i=1,n
         whts(i)=1
 1200 continue
c     
      do 1600 i=1,n
         do 1400 j=1,n
            if (i .ne. j) whts(i)=whts(i)/(ts(i)-ts(j))
 1400    continue
 1600 continue
c     
c     this uses the '2nd form' of barycentric interpolation, first
c     form the denominator
c     
      dd=0
      do 2000 i=1,n
         dd=dd+whts(i)/(ttt-ts(i))
 2000 continue
c     
c     and next the interpolation coefficients
c     
      do 2400 i=1,n
         coefs(i)=whts(i)/(ttt-ts(i))/dd
 2400 continue
c     
      return
      end



c**********************************************************************
      subroutine hpotgrad2d_sdp2(wavek,source,dipvec,
     $     target,ifpot,pot,potd,ifgrad,grad,gradd,ifhess,hess,hessd)
      implicit real *8 (a-h,o-z)
c**********************************************************************
c
c     This subroutine calculates the potential POT, field GRAD and
c     Hessian HESS at the target point TARGET, due to a *unit* charge at
C     SOURCE; and also
c     potential POTD, field GRADD, and Hessian HESSD at the same target
c     point, due to a *unit* dipole at SOURCE with unit direction DIPVEC.
c     The scaling is that required of the delta function response: i.e.,
c     
c               pot = H_0(k*r)*(i/4)
c               grad = gradient(pot)
c               hess = Hessian
c
c     This routine is used by lp2d curve tools, lpclosedcurve.f. The point
c     is that it uses only one hank103 call for H_0 and H_1 for everything.
c     Tweaked by A. Barnett, from hpotgrad2d_sdp of Z. Gimbutas.
c     Needs hank103 from HFMM2D library.
c
c---------------------------------------------------------------------
c     INPUT:
c
c     ifpot      : flag for computing potentials
c                                  ifpot = 0 -> don't compute 
c                                  ifpot = 1 -> do compute 
c     ifgrad     : flag for computing gradients
c                               ifgrad = 0 -> don't compute 
c                               ifgrad = 1 -> do compute 
c     ifhess     : flag for computing hessians
c                               ifhess = 0 -> don't compute 
c                               ifhess = 1 -> do compute 
c     source    : location of the source 
c     dipvec    : dipole orientation
c     target    : location of the target
c     wavek     : Helmholtz parameter
c
c---------------------------------------------------------------------
c     OUTPUT:
c
c     pot       : calculated potential due to monopole potential
c     grad      : calculated monopole gradient
c     hess      : calculated monopole hessian
c     potd      : calculated dipole potential
c     gradd     : calculated dipole gradient
c     hessd     : calculated dipole hessian
c
c---------------------------------------------------------------------
      dimension source(2),target(2)
      complex *16 wavek,pot,grad(2),hess(3), potd,gradd(2),hessd(3)
      complex *16 dipstr
      dimension dipvec(2)
      complex *16 z, h0, h1, h2, h3, h4, cd, h2z, zk, ima, ima4, ima4inv
      complex *16 hx,hy
      complex *16 hxx,hxy,hyy
      complex *16 hxxx,hxxy,hxyy,hyyy
c
      data ima/(0.0d0,1.0d0)/
c
c ... Calculate offsets and distance
c
      ima4=-4*ima
      ima4inv=ima/4
c
      xdiff=target(1)-source(1)
      ydiff=target(2)-source(2)
      rr=xdiff*xdiff+ydiff*ydiff
      r=sqrt(rr)
      z=wavek*r
c
      ifexpon = 1
      call hank103(z, h0, h1, ifexpon)
c
c
      if( ifpot.eq.1) pot = h0*ima4inv
c
      if (ifgrad.eq.1) then
         cd = -h1*(wavek*ima4inv/r)
         grad(1) = cd*xdiff
         grad(2) = cd*ydiff
c         ctheta=xdiff/r
c         stheta=ydiff/r
c         cd = -h1*(wavek*charge*ima4inv)
c         grad(1) = cd*ctheta
c         grad(2) = cd*stheta
      endif
c
      if (ifhess.eq.1) then
         cd = (wavek*ima4inv/r)/rr
         h2z=(-z*h0+2*h1)
         hess(1) = cd*(h2z*xdiff*xdiff-rr*h1)
         hess(2) = cd*(h2z*xdiff*ydiff      )
         hess(3) = cd*(h2z*ydiff*ydiff-rr*h1)
c         ctheta=xdiff/r
c         stheta=ydiff/r
c         cd = (wavek*charge*ima4inv/r)
c         h2z=(-z*h0+2*h1)
c         hess(1) = cd*(h2z*ctheta*ctheta-h1)
c         hess(2) = cd*(h2z*ctheta*stheta   )
c         hess(3) = cd*(h2z*stheta*stheta-h1)
      endif
c
c
      ctheta=xdiff/r
      stheta=ydiff/r

      h2 = 2*h1/z-h0
      h3 = 4*h2/z-h1
c
      cd = h1/r*wavek*ima4inv
      dotprod = xdiff*dipvec(1)+ydiff*dipvec(2)

      if(ifpot.eq.1) potd = cd*dotprod
c
      if (ifgrad.eq.1) then
         cd = -wavek**2*ima4inv
         hxx = cd*(h2*(ctheta*ctheta-0.5d0)-h0/2)
         hxy = cd*(h2*ctheta*stheta             )
         hyy = cd*(h2*(stheta*stheta-0.5d0)-h0/2)
         gradd(1) = hxx*dipvec(1)+hxy*dipvec(2)
         gradd(2) = hxy*dipvec(1)+hyy*dipvec(2)
      endif
c
      if (ifhess.eq.1) then
         cd = -wavek**3*ima4inv
         hxxx = cd*(
     $      -h1/2*(-1.5d0)
     $      -h3/2*(+ctheta*ctheta-0.5d0-stheta*stheta)
     $      )*ctheta
         hxxy = cd*(
     $      -h1/2*(-0.5d0)
     $      -h3/2*(1.5d0*ctheta*ctheta-0.5d0*stheta*stheta)
     $      )*stheta
         hxyy = cd*(
     $      -h1/2*(-0.5d0)
     $      -h3/2*(1.5d0*stheta*stheta-0.5d0*ctheta*ctheta)
     $      )*ctheta
         hyyy = cd*(
     $      -h1/2*(-1.5d0)
     $      -h3/2*(+stheta*stheta-0.5d0-ctheta*ctheta)
     $      )*stheta
         hessd(1) = hxxx*dipvec(1)+hxxy*dipvec(2)
         hessd(2) = hxxy*dipvec(1)+hxyy*dipvec(2)
         hessd(3) = hxyy*dipvec(1)+hyyy*dipvec(2)
      endif
c
      return
      end


c************************** debugging ****************************************
      subroutine dummy(n, u)
c     test the mwrap and calling from matlab. implicit is dangerous.
      implicit real *8 (a-h,o-z)
C     The following seems to set up output for stderr which appears in matlab:
      call prini(6,13)
C     test that claim...  (note prini call from matlab has no effect)
      call prinf('dummy n=*',n,1)
      u = 7.0D0
      call prin2('u=*',u,1)
      return
      end   

      subroutine testkernel(par1)
c     ---- debug kernel funcs for fixed source & target. See testapplyalpop.m
      implicit none
      real *8 src(2),snor(2),targ(2),tnor(2)
      complex *16 par1, par2, par3, f(4), grad, hess, ch
      call prini(6,13)
      src(1) = 0d0
      src(2) = 0d0
      targ(1) = 1d0
      targ(2) = 2d0
      snor(1) = 1d0
      snor(2) = 0d0
      tnor(1) = 0d0
      tnor(2) = -1d0
      par2 = (-1,0)
      par3 = (0,0)
      ch = (1,0)
c      call hpotgrad2d(0,0,src,ch,targ,par1,f,grad,hess)
      call fourkernels(src,snor,targ,tnor,par1,par2,par3,f)
      call prin2('all 4 kernels f=*',f,8)
      return
      end
      
